package de.halbekunst.juplo.cachecontrol;

import de.halbekunst.juplo.cachecontrol.AcceleratorFilter.AccelerationWrapper;
import de.halbekunst.juplo.cachecontrol.annotations.CacheSeconds;
import de.halbekunst.juplo.cachecontrol.annotations.Accepts;
import de.halbekunst.juplo.cachecontrol.annotations.AdditionalHeaders;
import de.halbekunst.juplo.cachecontrol.annotations.LastModified;
import de.halbekunst.juplo.cachecontrol.annotations.ETag;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Component;

/**
 *
 * @author kai
 */
@Component
public class CacheControl {
  private final static Logger log = LoggerFactory.getLogger(CacheControl.class);

  private static final ThreadLocal<CacheMethodHandle> tl = new ThreadLocal<CacheMethodHandle>();

  @Autowired @Qualifier("cacheSeconds") private Integer defaultCacheSeconds;
  @Autowired @Qualifier("lastModified") private Long defaultLastModified;


  public void init(CacheMethodHandle handle) {
    CacheControl.tl.set(handle);
  }

  void init(Object handler, AccelerationWrapper wrapper) throws NoSuchMethodException {
    CacheControl.tl.set(new ReflectionCacheMethodHandle(handler, wrapper == null ? false : wrapper.zipped));
  }

  public boolean decorate(
      HttpServletRequest request,
      HttpServletResponse response
      )
  {
    try {
    CacheMethodHandle handle = CacheControl.tl.get();

    /** Doppelte Ausführung verhindern... */
    if (handle == null) {
      /** Dekoration wurde bereits durchgeführt! */
      return true;
    }

    /**
     * Alle Antworten (insbesondere auch 304) sollen nach dem {@plainlink
     * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.18 RFC 2616,
     * Abschnitt 14.18} einen Date-Header enthalten
     */
    response.setDateHeader(Headers.HEADER_DATE, handle.getTimestamp());

    /** Besondere Maßnahmen für besondere HTTP-Status-Codes ?!? */
    int status = handle.accepts(request);
    switch (status) {
      case HttpServletResponse.SC_OK: // 200
      case HttpServletResponse.SC_NO_CONTENT: // 204
      case HttpServletResponse.SC_PARTIAL_CONTENT: // 206
        /** Normale Antwort! Antwort dekorieren... */
        break;
      case HttpServletResponse.SC_BAD_REQUEST: // 400
      case HttpServletResponse.SC_UNAUTHORIZED: // 401
      case HttpServletResponse.SC_PAYMENT_REQUIRED: // 402
      case HttpServletResponse.SC_FORBIDDEN: // 403
      case HttpServletResponse.SC_NOT_FOUND: // 404
      case HttpServletResponse.SC_METHOD_NOT_ALLOWED: // 405
      case HttpServletResponse.SC_NOT_ACCEPTABLE: // 406
      case HttpServletResponse.SC_PROXY_AUTHENTICATION_REQUIRED: // 407
      case HttpServletResponse.SC_REQUEST_TIMEOUT: // 408
      case HttpServletResponse.SC_CONFLICT: // 409
      case HttpServletResponse.SC_GONE: // 410
      case HttpServletResponse.SC_LENGTH_REQUIRED: // 411
      case HttpServletResponse.SC_PRECONDITION_FAILED: // 412
      case HttpServletResponse.SC_REQUEST_ENTITY_TOO_LARGE: // 413
      case HttpServletResponse.SC_REQUEST_URI_TOO_LONG: // 414
      case HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE: // 415
      case HttpServletResponse.SC_REQUESTED_RANGE_NOT_SATISFIABLE: // 416
      case HttpServletResponse.SC_INTERNAL_SERVER_ERROR: // 500
      case HttpServletResponse.SC_NOT_IMPLEMENTED: // 501
      case HttpServletResponse.SC_SERVICE_UNAVAILABLE: // 503
      case HttpServletResponse.SC_HTTP_VERSION_NOT_SUPPORTED: // 505
        return true;
      default:
        /**
         * Es ist nicht klar, was der Handler noch machen wird/muss:
         * Antwort nicht dekorieren und Kontroller an den Handler übergeben...
         */
        return false;
    }

    if (handle.isZipped())
      response.addHeader(Headers.HEADER_CONTENT_ENCODING, "gzip");

    Map<String,String> headers = handle.getAdditionalHeaders(request);
    for (String name : headers.keySet())
      response.addHeader(name, headers.get(name));

    String url = null;
    if (log.isDebugEnabled()) {
      if (request.getQueryString() == null) {
        url = request.getRequestURI();
      }
      else {
        StringBuilder builder = new StringBuilder();
        builder.append(request.getRequestURI());
        builder.append('?');
        builder.append(request.getQueryString());
        url = builder.toString();
      }
    }

    int cacheSeconds = handle.getCacheSeconds(request);
    if (cacheSeconds < 1) {
      log.debug("{}: caching disabled!", url);
      response.setDateHeader(Headers.HEADER_DATE, handle.getTimestamp());
      response.setDateHeader(Headers.HEADER_EXPIRES, 0);
      response.addHeader(Headers.HEADER_PRAGMA, "no-cache");
      response.addHeader(Headers.HEADER_CACHE_CONTROL, "private");
      response.addHeader(Headers.HEADER_CACHE_CONTROL, "no-cache");
      response.addHeader(Headers.HEADER_CACHE_CONTROL, "no-store");
      response.addHeader(Headers.HEADER_CACHE_CONTROL, "max-age=0");
      response.addHeader(Headers.HEADER_CACHE_CONTROL, "s-max-age=0");
      return true;
    }

    long ifModifiedSince = -1;
    try {
      ifModifiedSince = request.getDateHeader(Headers.HEADER_IF_MODIFIED_SINCE);
    }
    catch (Exception e) {
      log.error("Exception while fetching If-Modified-Since: {}", e);
    }

    long lastModified = handle.getLastModified(request);

    /**
     * Sicherstellen, dass der Wert keine Millisekunden enthält, da die
     * Zeitangabe aus dem Modified-Since-Header keine Millisekunden enthalten
     * kann und der Test unten dann stets fehlschlagen würde!
     */
    lastModified = lastModified - (lastModified % 1000);

    String ifNoneMatch = request.getHeader(Headers.HEADER_IF_NONE_MATCH);
    String eTag = handle.getETag(request);

    /**
     * 304-Antworten sollen nach dem {@plainlink
     * http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.3.5 RFC
     * 2616, Abschnitt 10.3.5} einen ETag-Header enthalten, wenn auch die
     * 200-Antwort einen enthalten hätte.
     */
    if (eTag != null) {
      StringBuilder builder = new StringBuilder();
      if (handle.isETagWeak())
        builder.append("W/");
      builder.append('"');
      builder.append(eTag);
      builder.append('"');
      response.setHeader(Headers.HEADER_ETAG, builder.toString());
    }


    if (ifModifiedSince >= lastModified && lastModified > 0) {
      /**
       * request.getDateHeader liefert die Zeit als long, oder -1, wenn der
       * Header nicht existiert. D.h., wenn "If-Modified-Since" nicht gesetzt
       * ist, wird die komplette Seite ausgeliefert.
       * Der zusätzliche Test, ob lastModified größer 0 ist, ist nötig, um
       * Fehler auszuschließen, wenn die Implementierung von Cachable
       * negative Werte für Last-Modified zurückliefert.
       */
      if (log.isDebugEnabled())
        log.debug("{}: Not modified since {}", url, new Date(ifModifiedSince));

      if (ifNoneMatch == null) {
        /** Neue Anfrage oder HTTP/1.0 Client! */
        log.debug("{}: ETag nicht gesetzt -> 304", url);
        response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
        return false;
      }
    }

    if (ifNoneMatch != null) {
      boolean weak = false;
      if (ifNoneMatch.startsWith("W/")) {
        weak = true;
        ifNoneMatch = ifNoneMatch.substring(3, ifNoneMatch.length() - 1);
      }
      else {
        ifNoneMatch = ifNoneMatch.substring(1, ifNoneMatch.length() - 1);
      }

      if (!weak || (request.getMethod().equals("GET") && request.getHeader(Headers.HEADER_RANGE) == null)) {
        /**
         * Die Gleichheit gilt nur, wenn die ETag's der Anfrage _und_ der
         * Antwort stark sind (starke Gleichheit!), oder wenn die Antwort nur
         * schwache Gleichheit fordert...
         */
        if (ifNoneMatch.equals(eTag) && (handle.isETagWeak() || !weak)) {
          log.debug("{}: ETag {} not changed -> 304 ", url, ifNoneMatch);
          response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
          return false;
        }
      }
      else {
        log.warn("{}: ignoring weak ETag W/\"{}\", because the request was no GET-request or the Range-Header was present!", url, ifNoneMatch);
      }
    }


    log.debug("{}: first up!", url);

    /** HTTP/1.1-Caching-Header richtig setzen!! */
    response.setDateHeader(Headers.HEADER_LAST_MODIFIED, lastModified);

    /** Cache-Control für HTTP/1.1-Clients generieren */
    Map<String, String> cacheControl = new TreeMap<String, String>();

    /**
     * Wenn eins JSESSIONID in der URL enthalten ist, darf die Anfrage nur vom
     * Browser gecached werden!
     */
    if (request.isRequestedSessionIdFromURL()) {
      cacheControl.put("private", null);
    }
    else {
      /**
       * Hier muss nicht geprüft werden, ob cacheSeconds > 0 gilt, da in diesem
       * Fall oben bereits No-Cache-Header generiert und <code>false</code>
       * zurückgeliefert werden!
       *
       * Den Wert als <code>max-age</code> zu den Schlüssel-Wert-Paaren für den
       * <code>Cache-Control</code>-Header hinzufügen und einen entsprechenden
       * <code>Expires</code>-Header für HTTP/1.0-Clients setzen.
       */
      cacheControl.put("max-age", Integer.toString(cacheSeconds));
      response.setDateHeader(Headers.HEADER_EXPIRES, (handle.getTimestamp() + (long) cacheSeconds * 1000));
    }

    /** Dem Handler die Gelegenheit geben, den Cache-Controll-Header anzupassen */
    handle.cacheControl(request, cacheControl);


    if (cacheControl.containsKey("private")) {
      /**
       * HTTP/1.0 Caches davon abhalten, die Ressource zu cachen (vgl.: RFC
       * 2616, {@plainlink
       * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.9.3
       * Abschnitt 14.9.3} und {@plainlink
       * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.32
       * Abschnitt 14.32})
       */
      response.setDateHeader(Headers.HEADER_EXPIRES, 0l);
      response.addHeader(Headers.HEADER_PRAGMA, "no-cache");
    }

    StringBuilder builder = new StringBuilder();
    for (Entry<String, String> entry : cacheControl.entrySet()) {
      builder.setLength(0);
      builder.append(entry.getKey());
      if (entry.getValue() != null) {
        builder.append('=');
        builder.append(entry.getValue());
      }
      response.addHeader(Headers.HEADER_CACHE_CONTROL, builder.toString());
    }

    return true;
    }
    finally {
      /**
       * Thread-Locale-Variable zurücksetzen, damit
       * 1.) ein doppelter Aufruf dieser Methode pro Request erkannt werden kann
       * 2.) der nächste Request nicht mit dem selben Handle weiterarbeitet
       */
      CacheControl.tl.set(null);
    }
  }

  public void release() {
    CacheControl.tl.set(null);
  }


  class ReflectionCacheMethodHandle implements CacheMethodHandle {

    private Object handler;
    private long now = System.currentTimeMillis();
    private Integer cacheSeconds;
    private Long lastModified;
    private String eTag;
    private Map<String,String> additionalHeaders;
    private Method acceptsMethod;
    private Method cacheSecondsMethod;
    private Method lastModifiedMethod;
    private Method eTagMethod;
    private Method cacheControlMethod;
    private Method additionalHeadersMethod;
    private boolean isAcceptsMethodDefined;
    private boolean isCacheSecondsMethodDefined;
    private boolean isLastModifiedMethodDefined;
    private boolean isETagMethodDefined;
    private boolean isCacheControlMethodDefined;
    private boolean isAdditionalHeadersMethodDefined;
    private boolean weak;
    private boolean zipped;


    ReflectionCacheMethodHandle(Object handler, boolean zipped) throws NoSuchMethodException {

      this.handler = handler;
      this.zipped = zipped;

      cacheSeconds = CacheControl.this.defaultCacheSeconds;
      lastModified = CacheControl.this.defaultLastModified;
      eTag = "";

      /** Class-Level-Annotations auslesen */
      for (Annotation annotation : handler.getClass().getAnnotations()) {
        if (annotation.annotationType().equals(CacheSeconds.class)) {
          cacheSeconds = ((CacheSeconds)annotation).value();
          isCacheSecondsMethodDefined = true;
          continue;
        }
        if (annotation.annotationType().equals(LastModified.class)) {
          lastModified = ((LastModified)annotation).value();
          if (lastModified < 1) {
            /**
             * Ein Last-Modified-Header wurde angefordert, aber es wurde kein
             * statischer Wert spezifiziert:
             * globalen statischen Default-Wert benutzen!
             */
            lastModified = defaultLastModified;
          }
          isLastModifiedMethodDefined = true;
          continue;
        }
        if (annotation.annotationType().equals(ETag.class)) {
          ETag eTagAnnotation = (ETag)annotation;
          eTag = eTagAnnotation.value();
          weak = eTagAnnotation.weak();
          isETagMethodDefined = true;
          continue;
        }
        if (annotation.annotationType().equals(AdditionalHeaders.class)) {
          AdditionalHeaders additionalHeadersAnnotation = (AdditionalHeaders)annotation;
          additionalHeaders = new HashMap<String,String>();
          for (String header : additionalHeadersAnnotation.value()) {
            int i = header.indexOf(':');
            if (i < 0) {
              log.error("invalid header: [{}]", header);
            }
            else {
              String name = header.substring(0,i).trim();
              String value = header.substring(i+1,header.length()).trim();
              additionalHeaders.put(name, value);
            }
          }
          isAdditionalHeadersMethodDefined = true;
          continue;
        }
      }

      /** Method-Level-Annotations auslesen */
      for (Method method : handler.getClass().getMethods()) {
        for (Annotation annotation : method.getAnnotations()) {
          if (annotation.annotationType().equals(Accepts.class)) {
            if (isAcceptsMethodDefined)
              throw new IllegalArgumentException("Die Annotation @Accept wurde in der Klasse " + handler.getClass().getSimpleName() + " mehrfach verwendet!");
            acceptsMethod = method;
            isAcceptsMethodDefined = true;
            continue;
          }
          if (annotation.annotationType().equals(CacheSeconds.class)) {
            if (isCacheSecondsMethodDefined)
              throw new IllegalArgumentException("Die Annotation @CacheSeconds wurde in der Klasse " + handler.getClass().getSimpleName() + " mehrfach verwendet!");
            cacheSecondsMethod = method;
            isCacheSecondsMethodDefined = true;
            continue;
          }
          if (annotation.annotationType().equals(LastModified.class)) {
            if (isLastModifiedMethodDefined)
              throw new IllegalArgumentException("Die Annotation @LastModified wurde in der Klasse " + handler.getClass().getSimpleName() + " mehrfach verwendet!");
            lastModifiedMethod = method;
            isLastModifiedMethodDefined = true;
            continue;
          }
          if (annotation.annotationType().equals(ETag.class)) {
            if (isETagMethodDefined)
              throw new IllegalArgumentException("Die Annotation @ETag wurde in der Klasse " + handler.getClass().getSimpleName() + " mehrfach verwendet!");
            eTagMethod = method;
            weak = ((ETag)annotation).weak();
            isETagMethodDefined = true;
            continue;
          }
          if (annotation.annotationType().equals(de.halbekunst.juplo.cachecontrol.annotations.CacheControl.class)) {
            if (isCacheControlMethodDefined)
              throw new IllegalArgumentException("Die Annotation @CacheControl wurde in der Klasse " + handler.getClass().getSimpleName() + " mehrfach verwendet!");
            cacheControlMethod = method;
            isCacheControlMethodDefined = true;
            continue;
          }
          if (annotation.annotationType().equals(AdditionalHeaders.class)) {
            if (isAdditionalHeadersMethodDefined)
              throw new IllegalArgumentException("Die Annotation @AdditionalHeaders wurde in der Klasse " + handler.getClass().getSimpleName() + " mehrfach verwendet!");
            additionalHeadersMethod = method;
            isAdditionalHeadersMethodDefined = true;
            continue;
          }
        }
      }

      if (!isAdditionalHeadersMethodDefined)
        additionalHeaders = new HashMap<String,String>();
    }


    @Override
    public boolean isZipped() {
      return zipped;
    }

    @Override
    public long getTimestamp() {
      return now;
    }

    @Override
    public int accepts(HttpServletRequest request) throws IllegalArgumentException {
      if (acceptsMethod == null) {
        return HttpServletResponse.SC_OK;
      }
      else {
        try {
          return (Integer)acceptsMethod.invoke(handler, request);
        }
        catch (Exception e) {
          throw new IllegalArgumentException(e);
        }
      }
    }

    @Override
    public int getCacheSeconds(HttpServletRequest request) throws IllegalArgumentException {
      if (cacheSecondsMethod == null) {
        return cacheSeconds;
      }
      else {
        try {
          return (Integer)cacheSecondsMethod.invoke(handler, request);
        }
        catch (Exception e) {
          throw new IllegalArgumentException(e);
        }
      }
    }

    @Override
    public long getLastModified(HttpServletRequest request) throws IllegalArgumentException {
      if (lastModifiedMethod == null) {
        return lastModified;
      }
      else {
        try {
          return (Long)lastModifiedMethod.invoke(handler, request);
        }
        catch (Exception e) {
          throw new IllegalArgumentException(e);
        }
      }
    }

    @Override
    public String getETag(HttpServletRequest request) throws IllegalArgumentException {
      if (eTagMethod == null) {
        return eTag;
      }
      else {
        try {
          return (String)eTagMethod.invoke(handler, request);
        }
        catch (Exception e) {
          throw new IllegalArgumentException(e);
        }
      }
    }

    @Override
    public boolean isETagWeak() {
      return weak;
    }

    @Override
    public void cacheControl(
        HttpServletRequest request,
        Map<String, String> cacheControlMap
        )
        throws IllegalArgumentException
    {
      if (cacheControlMethod != null) {
        try {
          cacheControlMethod.invoke(handler, request, cacheControlMap);
        }
        catch (Exception e) {
          throw new IllegalArgumentException(e);
        }
      }
    }

    @Override
    public Map<String,String> getAdditionalHeaders(HttpServletRequest request) throws IllegalArgumentException {
      if (additionalHeadersMethod == null) {
        return additionalHeaders;
      }
      else {
        try {
          return (Map<String,String>)additionalHeadersMethod.invoke(handler, request);
        }
        catch (Exception e) {
          throw new IllegalArgumentException(e);
        }
      }
    }
  }
}
